use std::{fs::File, io, sync::Arc};

use memmap::Mmap;

use super::BagReader;
use crate::{parse, Error, Result};

#[derive(Clone)]
pub struct MmapReader {
    mmap: Arc<Mmap>,
    pos: usize,
}

impl MmapReader {
    pub fn new(file: File) -> Result<Self> {
        // Safety: ¯\_(ツ)_/¯
        let mmap = unsafe { Mmap::map(&file) }?;
        let mmap = Arc::new(mmap);
        Ok(Self { mmap, pos: 0 })
    }
}

impl io::Read for MmapReader {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        if let Some(slice) = self
            .mmap
            .get(self.pos..)
            .and_then(|unread| unread.get(..buf.len()))
        {
            buf.copy_from_slice(slice);
            Ok(slice.len())
        } else {
            Ok(0)
        }
    }
}

impl BagReader for MmapReader {
    type Read = Self;

    fn as_read(&mut self) -> &mut Self::Read {
        self
    }

    fn read_parser<'a, O: 'a, P>(&'a mut self, mut parser: P) -> Result<O>
    where
        P: nom::Parser<&'a [u8], O, parse::Error<&'a [u8]>>,
    {
        let input = self.mmap.get(self.pos..).ok_or(Error::Eof)?;
        match parser.parse(input) {
            Ok((rest, output)) => {
                self.pos += input.len() - rest.len();
                Ok(output)
            }
            Err(nom::Err::Incomplete(_)) => Err(Error::Eof),
            Err(nom::Err::Error(e) | nom::Err::Failure(e)) => Err(e.into()),
        }
    }

    fn seek(&mut self, pos: io::SeekFrom) -> Result<()> {
        match pos {
            io::SeekFrom::Start(pos) => {
                self.pos = pos as usize;
            }
            io::SeekFrom::End(pos) => {
                self.pos = (self.mmap.len() as i64 + pos) as usize;
            }
            io::SeekFrom::Current(pos) => {
                self.pos = ((self.pos as i64) + pos) as usize;
            }
        }
        Ok(())
    }
}
